#! /usr/bin/python
#A module containing utility functions like date and string manipulation.
#
# Copyright or Copr. INRIA/RESO - Romaric GUILLIER, Pascale VICAT-BLANC PRIMET 
#
# Pascale VICAT-BLANC PRIMET - <pascale.primet@inria.fr>
# Romaric GUILLIER           - <romaric.guillier@ens-lyon.fr>
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# 
# For more information, see the file COPYING
#
# $Id: NXEUtils.py,v 1.6 2008-11-06 11:26:18 rguillie Exp $

import string
import datetime
import math
import re

import sys

import NXETypes 
from NXETypes import execStep
import NXELogging

def normalize_whitespace(text):
    "Remove redundant whitespace from a string"
    return ' '.join(text.split())

def split_comma(text):
    "Split text according to commas"
    return text.split(',')

def explicit_range(text):
    "Transform a token like xx-yy into a list [xx, xx+1,..,yy-1,yy]"
    number_list = []

    splitteddash = text.split('-')
    #print splitteddash
    
    if len(splitteddash) > 1:        
        number_list = range(int(splitteddash[0],10),int(splitteddash[1],10)+1)
    else:    
        number_list = [int(splitteddash[0],10)]
    
    return number_list 

def split_hostname(name):
    """Method to split a johndoe@node into a tuple containing the separated user (if empty, providing the current username)
       and hostname.
    """    
    import os

    splittedname = name.split('@')
    
    if len(splittedname) > 1:
        return (splittedname[0],splittedname[1])
    else: 
        return (os.getlogin(), splittedname[0]) 


def split_basename (name, number = -1):
    """Method to split the basename of a node. It returns a tuple containing the basename
    and a list of numbers corresponding to a
    ex: if name=sagittaire{1-5}, we should return (sagittaire, (1,2,3,4,5)) 
    the number parameter is used when there is an * at the end of the basename"""

    basename = ""
    numberlist = []

    #We got a 'valid star' when it is the last character of string
    stardetected = re.split( '\*$' , name) 
    if len(stardetected) > 1:
        basename = stardetected[0]
        numberlist = range(1,number+1)                          
        return (basename, numberlist)    

    #We can also indicate braces delimiting ranges separated by commas
    bracedetected = re.split( '\{(?P<range>.*)\}$', name)         
    if len(bracedetected) > 1:
        basename= bracedetected[0]                        
        splittedcomas = split_comma(bracedetected[1])
        #print splittedcomas
        for i in xrange(0, len(splittedcomas)):
            numberlist.extend(explicit_range(splittedcomas[i]))
        #print numberlist    
        return (basename, numberlist)

    #Normal case, just a name followed by any number
    normalcase = re.split( '(?P<range>[0-9]+$)', name)    
    #print normalcase
    basename = normalcase[0]
    #print basename
    numberlist = [int(normalcase[1],10)]
    #print numberlist
    
    return (basename, numberlist)

def offset_execStep(execStepList, offsetList, nodenumber):
    "A function that will add the given offsetList to the given execStepList"

    newExecStepList =[]

    if len(execStepList) > len(offsetList):
       #Great! Somebody has forgotten to put enough offsets in its list
       #Assuming that we have nothing but zeroes at the end
       sys.stderr.write("WARNING: padding offsets with zeroes due to mismatched lengths")
       for i in xrange( len(offsetList), len(execStepList) - 1):
           offsetList.append(0)

    #print "execStepList: " + str(len(execStepList))  

    for i in xrange(0, len(execStepList)):
       # print i
        if ((offsetList[i] == 0) or (nodenumber == 0)):
            newExecStepList.append(execStepList[i])
        else:
            execStepObj = execStepList[i]
            newExecStepList.append(execStep((execStepObj.date + (nodenumber * offsetList[i])), execStepObj.scriptname, execStepObj.scriptparams, execStepObj.label))

    return newExecStepList
    
def sort_execStep(execStepList):
    "A fonction that will sort by date the steps to execute in the right order"

    sortedList = []

    if len(execStepList) == 1:
        return

    #Putting the date in a "first column"
    for i in xrange(0, len(execStepList)):
        
        sortedList.append([execStepList[i].date, execStepList[i]])

    #Sorting using the first column (date) as an index
    sortedList.sort(key=lambda x:x[1])

    #Removing the first column
    for i in xrange(0, len(execStepList)):

        sortedList[i] = sortedList[i][1]

    execStepList = sortedList
    
    #for i in xrange(0, len(execStepList)):
    #    print "stepList" +  execStepList[i].toString()
        

def script_execStep(execStepList, filename, pid, location, timeout, target="localhost", extraoptions="", offset=0):
    """A script to write a small shell script that corresponds to the
       execution of several steps
    """

    curTimer=0

    shellfile = open(filename,'w')

    #Prologue
    shellfile.write("#!/bin/bash\n#Automatically generated script for communication between " + location + " and " + target + "\n\n")

    
    for i in xrange(0, len(execStepList)):

        shellfile.write("#Step " + str(i) + "\n")
        timeToSleep = execStepList[i].date - curTimer - offset
        if timeToSleep > 0:
            shellfile.write("sleep " + str(timeToSleep) + "\n")

        timeoutoption = ""
        if execStepList[i].timeout > 0:
            timeoutoption = " --timeout " + str(execStepList[i].timeout)
        else:
            timeoutoption = " --timeout " + str(timeout)
        
        commandstring = execStepList[i].scriptname + " --pid " + str(pid) + " --date " + str(execStepList[i].date) + " --target " + target + extraoptions + timeoutoption + " " + execStepList[i].scriptparams + "\n"

        shellfile.write(commandstring)

        curTimer = execStepList[i].date


    shellfile.write("#All finished\n")
    shellfile.flush()
    shellfile.close()



def getTime():
    "A function that returns the current time"
    return str(datetime.datetime.now())


def divList(listA, listB):
    """A function to divide listA[i] by listB[i] and returns the corresponding 
       list. Only the size of the first list matters.

       if ListB[i} == 0, we put a NaN in the table. It is probably not the best
       way to handle this, but at least we try to return a list that has the
       same length as the input.
    """
    tab = []
    for i in xrange(0,len(listA)):
    	if listB[i] != 0:
    	   tab.append(float(listA[i])/float(listB[i]))
	else:
	   tab.append(float("NaN"))	

    return tab

def cumulmodification(inputfilename, outputfilename, modtab = None, addtab = None , multab = None):  
    """A sightly more generic version of the diff(input, output) below
       except that it will do something like

       (line(n) + modtab * line(n-1) + addtab)*multab

       except that we need to do it column by column because the operation on
       table are limited.
    """
    
    inputfile = open(inputfilename, 'r') 
    linelist  = inputfile.readlines()
    inputfile.close()

    if (len(linelist) < 1):
        NXELogging.logging.printErr("Empty File " + inputfilename + "  aborting")
        return

    #initial check that all the tabs have the right size
    #We don't care if the tab are larger than what we have in the file
    colnumber=len(normalize_whitespace(linelist[0]).split(' '))

    #Setting the default value in case the arguments were
    if modtab is None:
        modtab = [None,] * colnumber
        for i in xrange(0,colnumber):
            modtab[i]=0.0

    if addtab is None:
        addtab = [None,] * colnumber
        for i in xrange(0,colnumber):
            addtab[i]=0.0

    if multab is None:
        multab = [None,] * colnumber
        for i in xrange(0,colnumber):
            multab[i] = 1.0

    if ( (len(modtab) < colnumber) or (len(addtab) < colnumber) or (len(multab) < colnumber)):
        NXELogging.logging.printErr("{Mod|add|mul}tab is too short: ", repr(modtab), repr(addtab), repr(multab))
        return


    outputfile = open(outputfilename, 'w')
    splittedpreviousline = normalize_whitespace(linelist[0]).split(' ')

    for lineindex in xrange(1, len(linelist)):

        outputstring = ""

        currentline=linelist[lineindex]
        splittedcurrentline = normalize_whitespace(currentline).split(' ')
        

        for columnindex in xrange(0, colnumber):            
            value = (float(splittedcurrentline[columnindex]) + modtab[columnindex] * float(splittedpreviousline[columnindex]) + addtab[columnindex])*multab[columnindex]

            if (columnindex == 0):
                outputstring = str(value)
            else:
                outputstring += " " + str(value)

        splittedpreviousline=splittedcurrentline

        outputfile.write(outputstring + "\n")

    outputfile.close()


def diff(inputfilename, ouputfilename):
    """A function that takes a file containing two columns of float and
       fill the second file with the same first column but the second one is the diff
       between two succesive values in the second column

       Yes! I'm not good at explaining this kind of things, but read the code, it is obvious
       """

    inputfile = open(inputfilename, 'r')
    linelist  = inputfile.readlines()
    inputfile.close()


    outputfile = open(outputfilename, 'w')
    
     
    previousdate = 0.0
    firstlocaldate = 0.0

    for index in xrange(0, len(linelist)):

        currentline=linelist[index]
        splittedcurrentline = normalize_whitespace(currentline).split(' ')

        localdate=float(splittedcurrentline[0])
        distantdate=float(splittedcurrentline[1])
        
        diffdate = 0.0    
        
        if (previousdate == 0.0): 
            diffdate = 0.0
            relativedate = 0.0
            firstlocaldate = localdate
        else:
            diffdate = distantdate - previousdate
            relativedate = localdate - firstlocaldate
           
        previousdate = distantdate
        outputfile.write(str(relativedate) + " " + str(diffdate) + "\n")


    outputfile.close()

###
def printCompletionTabToFile(completiontab, outputfilename, tagtab = []):
    """A function that prints whatever is in the completion tab in a file
    Perhaps we should put also the part where we are putting the completion times
    in bins rather than just outputting to a stupid file
    """

    outputfile = open(outputfilename, 'w')

    if (len(tagtab) > 0):
        outputfile.write("#")
        for i in xrange(0,len(tagtab)):
            outputfile.write(tagtab[i] + "\t")
        outputfile.write("\n")

    for i in xrange(0, len(completiontab)):

        size, time = completiontab[i] 
        outputfile.write(str(size) + " " + str(time) + "\n")

    outputfile.close()

###
def printResultTabToFile(resulttab, outputfilename, maxtime, tagtab = []): 
     """A function that prints whatever is in the resultab line by line 
        It is assumed that what is in the resulttab is something like the
        output of the parse_iperf_result (a big string with \n in each slot)
        The tagtab is there if we need to put something pretty in the header
        
     """

     outputfile = open(outputfilename, 'w')

     if (len(tagtab) != len(resulttab)):
         NXELogging.logging.printOut("Your tagtab side is not matching the size of the resulttab",NXELogging.LOG_WARN)

     if (len(tagtab) > 0):
         outputfile.write("#")
         for i in xrange(0,len(tagtab)):
             outputfile.write(tagtab[i] + "\t")
         outputfile.write("\n")

     #We have at least the tagtab inside the file

     #Crazy optimisation
     if (len(resulttab) == 1):               
        outputfile.write(resulttab[0])
        outputfile.close()
        return

     temptab = [None, ] * len(resulttab)
     for i in xrange(0,len(resulttab)):
         temptab[i] = resulttab[i].split("\n") 
         
     #print int( 2 * maxtime)
     
     for i in xrange(0, int( 2 * maxtime) + 1):  
         outputstring = ""
         
         for j in xrange(0, len(resulttab)):
             #print i,j, len(temptab[j])
             if (len(temptab[j]) <= i  ):
                 #The case where we need to pad the file
                 outputstring += str(float(i)/2) + " 0.0 0.0"
             else:
                 outputstring += temptab[j][i]


             if (j < len(resulttab) - 1):
                 outputstring += " "
             else:
                 outputstring += "\n"
             

         outputfile.write(outputstring)

     
     outputfile.close()

# def printFileToTab(inputfilename, resulttab, tagtab):
#     """A function that for a strange reasons is some sort of reverse
#        of the printTabto

#        It will append the big string corresponding to the file in the next slot 
#        in the result tab. The first line will be parsed, just in case a description
#        was put on the first line.
#     """

#     inputfile = open(inputfilename, 'r')

#     inputtab = inputfilename.readlines()
#     outputstring = ""

#     if (inputtab[index][0] == "#"):
#         #if there is a # on the first line, it is a
#         tagline = normalize_whitespace(inputtab[index].replace('#',''))
#         tagtab.append(tagline)

#     for index in xrange(0,len(inputtab)):    
#         outputstring += inputtab[index] + "\n"
        
#     resulttab.append(outputstring)

#     inputfile.close()

def fillTabFromString(inputstring):
    """Do I have a good reason to do just anything more than tab = input.split ?
       Yes ! You need float !
    """

    splittedstring = inputstring.split(':')
    tab = [None,] * len(splittedstring)

    for i in xrange(0,len(splittedstring)):
        tab[i]=int(round(float(splittedstring[i])))    

    return tab

def halfround(number):
    """A function to wrap a number between  [n], [n]+.5 and [n]+1 """

    roundnumber = round(number)
    digit = number % 1
    
    if (digit < 0.25):
        return roundnumber
    elif ((digit >= 0.25) and (digit < 0.75)):
        return roundnumber + 0.5
    else:
        return roundnumber + 1


def rangefloat(start, stop, step):
    """Something to replace the range over floats, as I don't have a way to check that
       I have a better way from the Python library to do exactly the same thing
       If not think about making something that is iterable, otherwise, it is a bit
       ugly to use
    """

    count = int(math.ceil((stop - start) / step))
    #print count, stop, start, step
    if (count < 1):
        return []
    
    tab = [None,] * count

    tab[0] = start
    for i in xrange(1,count): 
        tab[i] = tab[i-1] + step


    return tab


if __name__ == '__main__':


    from sys import argv
    if len(argv) == 2:
        inputfilename = argv[1]
        outputfilename = "parsed_" + inputfilename
    elif len(argv) >= 3:
        inputfilename = argv[1]
        outputfilename = argv[2]

    diff(inputfilename, outputfilename)
    

# if __name__ == '__main__':

#     split_basename('pouet*',10 )
#     split_basename('coincoin{1-5,6,7-9}')
#     split_basename('pou*et*')
#     split_basename('pouet24')

